/* Copyright Amit P. Singh, Douglas L. Brutlag, Stanford University. */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "defines.h"
#ifdef DISPLAYON
#include <GL/gl.h>
#include <tcl.h>
#include <tk.h>
#include "display.h"
#endif

int globalgotone = 0;
int globalgottwo = 0;
int mypid;
char* lockdir;
char* dsspprogram;

extern void usageError(); 


#ifdef DISPLAYON

extern void setupGLWindow(char*);
extern int My_Tk_Init(int, char**, Tcl_AppInitProc*);
extern void InitPosition(int, int, int, int);
extern void addReshapeFunc(void (*Func)(GLsizei, GLsizei));
extern void addKeyFunc(int key, void (*Func)(void));
extern void addMouseFunc(int mouse, int mode, void (*Func)(AUX_EVENTREC *));
extern void tellTcl(int, char*, int);
extern void runTclFile(char*);
extern void tellTclAnything(char*);

extern Display *mydisplay;
extern Window GLwin;


GLfloat colorTable[2][3] = {
  {0.0, 0.0, 1.0},
  {1.0, 0.0, 0.0}
};

int currentColor = 0;

static GLfloat LdownXold = -1;
static GLfloat LdownYold = -1;
static GLfloat MdownXold = -1;
static GLfloat MdownYold = -1;
static GLfloat RdownXold = -1;
static GLfloat RdownYold = -1;

int jointModelNumber;
GLfloat currentTransformMat[16];
GLfloat jointModelMat[16];

int displayType = 0;

void createObjList(OBJ*);
void createJointList();
void display(void);
void init(void);
void myReshape(GLsizei, GLsizei);
void setBindings(void);
void setCurrentOne(int);
void setCurrentTwo(int);
void centerObjects();
void setLocalOrGlobal(int);
void setDisplayType(int);
void setDistThreshold(float);
void setThresholdVar(int);
void mouseLeftDown(AUX_EVENTREC*);
void mouseLeftUp(AUX_EVENTREC*);
void mouseMiddleDown(AUX_EVENTREC*);
void mouseMiddleUp(AUX_EVENTREC*);
void mouseRightDown(AUX_EVENTREC*);
void mouseRightUp(AUX_EVENTREC*);
void keyUP();
void keyDOWN();

#endif



OBJ* currentObjList[2];
int numCurrentObjs = 0;
OBJ* ObjList[2];

OBJ* currentObjList0[2];
OBJ* ObjList0[2];


extern char username[2][300];
extern char pdbname[2][10];
extern char chain[2]; 
extern FILE* filePtrs[2]; 
extern int newfile[2];
extern char outputDir[300];

int globalStartDisplay = 1;
int globalComputeCore = 1;

GLfloat distThreshold = 3.0;
int thresholdOFF = 0;
int displayPoints = 0;
FILE *outfile = NULL;

float globalRMSD = 0.0;
int globalNumAligned = 0;
float globalCoreRMSD = 0.0;
int globalCoreNumAligned = 0;
int  globalLastAlignment = 0;  /* 0 for structalign and 1 for refinealign */

float globalRMSD0 = 0.0;
int globalNumAligned0 = 0;
float globalCoreRMSD0 = 0.0;
int globalCoreNumAligned0 = 0;

int globalSearchDatabase = 0;
FILE *globalDatabaseFile = NULL;

extern void init_pairs();
extern void add_pair(Vector, Vector, float);
extern void match_pairs(Matrix, Vector);
extern void initsequences(OBJ *, OBJ *);
extern void align();
extern void report();
extern void initsequences_ss(OBJ *, OBJ *);
extern void align_ss(int);
extern void report_ss(int);
extern OBJ* readFile(int); 
extern void checkopts(int, char**);

extern int local;
extern int globalCalledAddPair;


void structAlignSingle();
void structAlignFull();
void alignUsingVectorsSingle();
void alignUsingVectorsFull();
void initAlignUsingVectorsSingle();
void initAlignUsingVectorsFull();
void sequenceAlign();
void quitFunction();
void printOutput();
void printStructAlignment();
void printAlignmentMatrices();
void loadAlignmentMatrices();
void coreAlignFull();
void coreAlignSingle();
void doAllAlignments();

void applyMat(OBJ*, OBJ*, GLfloat*);
void applyMtoV(GLfloat*, GLfloat*, GLfloat*);
void applyRottoV(GLfloat*, GLfloat*, GLfloat*);
void printMat(GLfloat*);
GLfloat findNearest(OBJ*, GLfloat*, char, GLfloat*, int*);
float computeRMSD(OBJ*, OBJ*);
float structAlign(OBJ*, OBJ*, int, float);
float alignUsingVectors(OBJ*, OBJ*, int, float, int);
float refineAlign(OBJ*, OBJ*, int, float, int);
void findBestSequence(int*, float*, int, int);    
void freeObj(OBJ*);

int inputObject(int);
void setupOutputFile();



/****************************************************************************************
                                  Matrix operations etc  
****************************************************************************************/

void myMatMult(GLfloat *m1, GLfloat*m2) {

  int i;
  GLfloat t[3];

  for (i = 0; i < 4; i++) {
    t[0] = m1[i*4] * m2[0] + m1[i*4 + 1] * m2[4] + m1[i*4 + 2] * m2[8];
    t[1] = m1[i*4] * m2[1] + m1[i*4 + 1] * m2[5] + m1[i*4 + 2] * m2[9];
    t[2] = m1[i*4] * m2[2] + m1[i*4 + 1] * m2[6] + m1[i*4 + 2] * m2[10];

    m1[i*4] = t[0];
    m1[i*4 + 1] = t[1];
    m1[i*4 + 2] = t[2];

  }

  m1[12] = m1[12] + m2[12];
  m1[13] = m1[13] + m2[13];
  m1[14] = m1[14] + m2[14];

}



void applyMat(OBJ* retObj, OBJ* myobj, GLfloat* mat) {

  int i;
  float t[3];

  for (i = 0; i < myobj->numPts; i++) {
    retObj->points[i][0] = (myobj->points[i][0]) * mat[0] + 
                           (myobj->points[i][1]) * mat[4] +
                           (myobj->points[i][2]) * mat[8] + mat[12];
    retObj->points[i][1] = (myobj->points[i][0]) * mat[1] + 
                           (myobj->points[i][1]) * mat[5] +
                           (myobj->points[i][2]) * mat[9] + mat[13];
    retObj->points[i][2] = (myobj->points[i][0]) * mat[2] + 
                           (myobj->points[i][1]) * mat[6] +
                           (myobj->points[i][2]) * mat[10] + mat[14];

  }
}



void applyMtoV(GLfloat *vout, GLfloat *vin, GLfloat *matin) {

  GLfloat t[3];

  t[0] = (vin[0])* matin[0] + (vin[1])* matin[4] + (vin[2])* matin[8] + matin[12];
  t[1] = (vin[0])* matin[1] + (vin[1])* matin[5] + (vin[2])* matin[9] + matin[13];
  t[2] = (vin[0])* matin[2] + (vin[1])* matin[6] + (vin[2])* matin[10] + matin[14];

  vout[0] = t[0];
  vout[1] = t[1];
  vout[2] = t[2];

}


void applyRottoV(GLfloat *vout, GLfloat *vin, GLfloat *matin) {

  GLfloat t[3];

  t[0] = (vin[0])* matin[0] + (vin[1])* matin[4] + (vin[2])* matin[8];
  t[1] = (vin[0])* matin[1] + (vin[1])* matin[5] + (vin[2])* matin[9];
  t[2] = (vin[0])* matin[2] + (vin[1])* matin[6] + (vin[2])* matin[10];

  vout[0] = t[0];
  vout[1] = t[1];
  vout[2] = t[2];

}



float computeRMSD(OBJ* query, OBJ* target) {

  GLfloat tempVec[3];
  GLfloat tdist;
  GLfloat rmsd = 0.0;
  int i, count = 0, tnum;

  for (i = 0; i < target->numPts; i++) {
    tdist = findNearest(query, target->points[i], target->sstype[i], tempVec, &tnum);
    if ((tdist <= distThreshold) || (thresholdOFF)) {
      rmsd += tdist*tdist;
      count++;
    }    
  }

  rmsd = sqrt((rmsd)/((GLfloat)(count)));
  globalNumAligned = count;

  return rmsd;
}


float distBetweenVecs(float *vec0, float *vec1) {

  return sqrt((vec1[0] - vec0[0])*(vec1[0] - vec0[0]) +
	      (vec1[1] - vec0[1])*(vec1[1] - vec0[1]) +
	      (vec1[2] - vec0[2])*(vec1[2] - vec0[2]));

}


float distanceBetweenPoints(float *p1, float *p2) {

  return sqrt((p1[0] - p2[0])*(p1[0] - p2[0]) +
	      (p1[1] - p2[1])*(p1[1] - p2[1]) +
	      (p1[2] - p2[2])*(p1[2] - p2[2]));
}
  

GLfloat findNearest(OBJ* target, GLfloat* pVec, char ss, GLfloat* rVec, int *targetNum) {

  int i, num = 0;
  GLfloat min = 10000.0;
  GLfloat dist;

  for (i = 0; i < target->numPts; i++) {
    
    if ((ss == target->sstype[i]) || (ss == 'U') || (target->sstype[i] == 'U')) {
      
      dist = sqrt((target->points[i][0] - pVec[0])*(target->points[i][0] - pVec[0]) +
		  (target->points[i][1] - pVec[1])*(target->points[i][1] - pVec[1]) +
		  (target->points[i][2] - pVec[2])*(target->points[i][2] - pVec[2]));
      
      if (dist < min) {
	min = dist;
	num = i;
      }
    }
  }

  rVec[0] = target->points[num][0];
  rVec[1] = target->points[num][1];
  rVec[2] = target->points[num][2];

  *targetNum = num;

  return min;

}

void printMat(GLfloat *m) {
  
  int i;

  for (i = 0; i<16; i++) {
    if (i%4 == 0) 
      fprintf(outfile, "\n");
    fprintf (outfile, "%f  ", m[i]);
  }

  fprintf(outfile, "\n");  
  
}



#ifdef DISPLAYON

/****************************************************************************************
                                  OpenGL stuff   
****************************************************************************************/


void createObjList(OBJ* myobject) {

  int i, j, count, start, end;
  char currentss;

  glNewList(myobject->modelNumber, GL_COMPILE);

  if (displayType == 0) {
    
    glLineWidth(2.0);
    glColor3f(myobject->color[0],myobject->color[1], myobject->color[2]);   

    glBegin(GL_LINE_STRIP);
    for (i = 0; i < myobject->numPts; i++)
      glVertex3fv(&(myobject->points[i][0]));
    glEnd(); 
  }

  else if (displayType == 1) {
    
    glLineWidth(2.0);
    
    for (i = 0; i < myobject->numPts; i++) {
      currentss = myobject->sstype[i];    
      if (currentss == 'H') {
	glColor3f(myobject->color[0],myobject->color[1] + 1.0, myobject->color[2]);
	count = 0;
	while ((i+count < myobject->numPts) && (currentss == myobject->sstype[i+count])){
	  count++;
	}
	
	glBegin(GL_LINE_STRIP);
	
	for (j = i; j < i+count; j++) 
	  glVertex3fv(&(myobject->points[j][0]));
	
	glEnd();
	
	i = i + count - 1;
	
      }
      else if (currentss == 'S') {
	glColor3f(myobject->color[0],myobject->color[1] + 0.5, myobject->color[2]);
	count = 0;
	while ((i+count < myobject->numPts) && (currentss == myobject->sstype[i+count])){
	  count++;
	}
	
	glBegin(GL_LINE_STRIP);
	
	for (j = i; j < i+count; j++) 
	  glVertex3fv(&(myobject->points[j][0]));
	
	glEnd();
	
	i = i + count - 1;
	
      }
      else if (currentss == 'U') {
	glColor3f(myobject->color[0],myobject->color[1], myobject->color[2]);
	count = 0;
	while ((i+count < myobject->numPts) && (currentss == myobject->sstype[i+count])){
	  count++;
	}
	
	start = i;
	end = i + count;
	
	if (start > 0) {
	  start--;
	}
	
	if (end < myobject->numPts - 1) {
	  end++;
	}
	
	glBegin(GL_LINE_STRIP);
	
	for (j = start; j < end; j++) 
	  glVertex3fv(&(myobject->points[j][0]));
	
	glEnd();
	
	
	i = i + count - 1;
	
      }
    }      
    
  }
  
  else {
    if ((displayType == 2) || (displayType == 4)) {
      
      glLineWidth(2.0);
      glColor3f(myobject->color[0],myobject->color[1], myobject->color[2]);

      for (i = 0; i < myobject->numPts; i++) {
	currentss = myobject->sstype[i];    
	if (currentss == 'H') {
	  count = 0;
	  while ((i+count < myobject->numPts) && (currentss == myobject->sstype[i+count])){
	    count++;
	  }
	  
	  glBegin(GL_LINE_STRIP);
	  
	  for (j = i; j < i+count; j++) 
	    glVertex3fv(&(myobject->points[j][0]));
	  
	  glEnd();
	  
	  i = i + count - 1;
	  
	}
      }
    }
      
    if ((displayType == 3) || (displayType == 4)) {
      glLineWidth(2.0);
      glColor3f(myobject->color[0],myobject->color[1], myobject->color[2]);
      
      for (i = 0; i < myobject->numPts; i++) {
	currentss = myobject->sstype[i];    
	if (currentss == 'S') {
	  count = 0;
	  while ((i+count < myobject->numPts) && (currentss == myobject->sstype[i+count])){
	    count++;
	  }
	  
	  glBegin(GL_LINE_STRIP);
	  
	  for (j = i; j < i+count; j++) 
	    glVertex3fv(&(myobject->points[j][0]));
	  
	  glEnd();
	  
	  i = i + count - 1;
	  
	}    
      }
    }

    if (displayType == 5) {
      glLineWidth(2.0);
      glColor3f(myobject->color[0],myobject->color[1], myobject->color[2]);
      for (i = 0; i < myobject->numPts; i++) {
	currentss = myobject->sstype[i];    
	if (currentss == 'U') {
	  count = 0;
	  while ((i+count < myobject->numPts) && (currentss == myobject->sstype[i+count])){
	    count++;
	  }
	  
	  start = i;
	  end = i + count;
	  
	  if (start > 0) {
	    start--;
	  }
	  
	  if (end < myobject->numPts - 1) {
	    end++;
	  }
	  
	  glBegin(GL_LINE_STRIP);
	  
	  for (j = start; j < end; j++) 
	    glVertex3fv(&(myobject->points[j][0]));
	  
	  glEnd();
	  
	  
	  i = i + count - 1;
	  
	}
      }      
    }
  }
  
  glEndList();

}

  
void createJointList() {
  
  int i;
  GLfloat tempvec[3];

  glNewList(jointModelNumber, GL_COMPILE);

  glLineWidth(1.0);
  glColor3f(1.0,1.0,1.0);
  
  glBegin(GL_LINES);

  for (i = 0; i < ObjList[0]->numPts; i++) {
    if (ObjList[0]->pairNums[i] != -1) {
      applyMtoV(tempvec, ObjList[0]->points[i], ObjList[0]->objMat);
      glVertex3fv(tempvec);
      applyMtoV(tempvec, ObjList[1]->points[(ObjList[0]->pairNums[i])], ObjList[1]->objMat);
      glVertex3fv(tempvec);
      
    }
  }

  glEnd();

  glEndList();

  glPushMatrix();
  glLoadIdentity();
  glGetFloatv(GL_MODELVIEW_MATRIX, jointModelMat);
  glPopMatrix();

  displayPoints = 1;

}
    

void display(void) {
  
  int i;

  if (globalStartDisplay) {
    glClearColor(0.0, 0.0, 0.0, 1.0); 
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 
    
    if (globalSearchDatabase == 0) {
      if (globalgotone) {
	glPushMatrix();
	glMultMatrixf(ObjList[0]->objMat);
	glCallList(ObjList[0]->modelNumber);
	glPopMatrix();
      }      
      if (globalgottwo) {
	glPushMatrix();
	glMultMatrixf(ObjList[1]->objMat);
	glCallList(ObjList[1]->modelNumber);
	glPopMatrix();
      }      
      
#if DRAWJOINTS
      if (displayPoints == 1) {
	glPushMatrix();
	glMultMatrixf(jointModelMat);      
	glCallList(jointModelNumber);
	glPopMatrix();
      }
#endif
      
    }
    glFlush();
    glXSwapBuffers(mydisplay, GLwin); 
  }

}

void init(void) {
  
  int i;

  glMatrixMode(GL_MODELVIEW); 
  glShadeModel(GL_FLAT);
  glColor3f(1.0, 0.0, 0.0);

  glLoadIdentity();
  glTranslatef(EYEX, EYEY, EYEZ); 

  glFlush();

}

void myReshape(GLsizei w, GLsizei h) {

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-SIZEX/2.0, SIZEX/2.0, -SIZEY/2.0, SIZEY/2.0, -90.0, 90.0);
  /*  glFrustum(-SIZEX/2.0, SIZEX/2.0, -SIZEY/2.0, SIZEY/2.0, 20.0, 90.0);*/
  glMatrixMode(GL_MODELVIEW); 
  glViewport(0, 0, w, h); 

  display();

}




/****************************************************************************************
                            User interface(tcl-tk) and graphics events
****************************************************************************************/

void setBindings(void) {

  addKeyFunc(AUX_UP, keyUP);      /* increase threshold */
  addKeyFunc(AUX_DOWN, keyDOWN);  /* decrease threshold */

  addMouseFunc(AUX_LEFTBUTTON, AUX_MOUSELOC, mouseLeftDown);
  addMouseFunc(AUX_LEFTBUTTON, AUX_MOUSEUP, mouseLeftUp);
  addMouseFunc(AUX_MIDDLEBUTTON, AUX_MOUSELOC, mouseMiddleDown);
  addMouseFunc(AUX_MIDDLEBUTTON, AUX_MOUSEUP, mouseMiddleUp);
  addMouseFunc(AUX_RIGHTBUTTON, AUX_MOUSELOC, mouseRightDown);
  addMouseFunc(AUX_RIGHTBUTTON, AUX_MOUSEUP, mouseRightUp);
}


void setCurrentOne(int val) {

  if (val == 1) {
      
    if (numCurrentObjs == 0) {     
      currentObjList[0] = ObjList[0];
      numCurrentObjs = 1;
    }
    else if ((numCurrentObjs == 1) && (currentObjList[0] == ObjList[1])) {
      currentObjList[1] = ObjList[0];
      numCurrentObjs = 2;
    }
  }
  
  else if (val == 0) {
    
    if (numCurrentObjs == 1) {
      numCurrentObjs = 0;
    }
    
    else if (numCurrentObjs == 2) {
      currentObjList[0] = ObjList[1];
      numCurrentObjs = 1;
    }
  }
  
}

void setCurrentTwo(int val) {

  if (val == 1) {

    if (numCurrentObjs == 0) {     
      currentObjList[0] = ObjList[1];
      numCurrentObjs = 1;
    }
    else if ((numCurrentObjs == 1) && (currentObjList[0] == ObjList[0])) {
      currentObjList[1] = ObjList[1];
      numCurrentObjs = 2;
    }
  }
  
  else if (val == 0) {
    
    if (numCurrentObjs == 1) {
      numCurrentObjs = 0;
    }
    
    else if (numCurrentObjs == 2) {
      currentObjList[0] = ObjList[0];
      numCurrentObjs = 1;
    }
  }

}


void centerObjects() {

  int i;

  for (i = 0; i < numCurrentObjs; i++) {
    
    glPushMatrix();
    
    glLoadIdentity();
    glTranslatef(-currentObjList[i]->center[0], -currentObjList[i]->center[1],  
		 -currentObjList[i]->center[2]);
    glMultMatrixf(currentObjList[i]->objMat);
    glGetFloatv(GL_MODELVIEW_MATRIX, currentObjList[i]->objMat);

    currentObjList[i]->center[0] = 0.0;
    currentObjList[i]->center[1] = 0.0;
    currentObjList[i]->center[2] = 0.0;
    
    glPopMatrix();
    
  }
  
  displayPoints = 0;
  display();

}



void setLocalOrGlobal(int i) {

  local = i;

}


void setDisplayType(int i) {
  
  displayType = i;
  createObjList(ObjList[0]);
  createObjList(ObjList[1]);
  
  display();
  
}

void setDistThreshold(float i) {
  
  distThreshold = i;

}

void setThresholdVar(int i) {

  if (i == 1) 
    thresholdOFF = 0;
  else if (i == 0)
    thresholdOFF = 1;

}

void mouseLeftDown(AUX_EVENTREC *mouseEvent) {

  GLfloat Xval, Yval, Xt, Yt;
  GLfloat transValue[3];
  int i;
  int smallestObjLen = 100000;
  int smallestObjNum = -1;
  

  Xt = (float)(mouseEvent->data[AUX_MOUSEX]);
  Yt = (float)(mouseEvent->data[AUX_MOUSEY]);

  if ((LdownXold == -1.0) && (LdownYold == -1.0)) {
    LdownXold = Xt;
    LdownYold = Yt;
    if (numCurrentObjs != 2) 
      displayPoints = 0;
  }
  else {

    Xval = (Xt - LdownXold)*0.5;
    Yval = (Yt - LdownYold)*0.5;

    for (i = 0; i < numCurrentObjs; i++) {
      if (currentObjList[i]->numPts < smallestObjLen) {
	smallestObjLen = currentObjList[i]->numPts;
	smallestObjNum = i;
      }
    }


    if (smallestObjNum != -1) {
      transValue[0] = currentObjList[smallestObjNum]->center[0];
      transValue[1] = currentObjList[smallestObjNum]->center[1];
      transValue[2] = currentObjList[smallestObjNum]->center[2];

    }
    
    glPushMatrix();
    
    glLoadIdentity();
    glTranslatef(transValue[0], transValue[1], transValue[2]);
    glRotatef(Xval, 0.0, 1.0, 0.0);
    glRotatef(Yval, 1.0, 0.0, 0.0);
    glTranslatef(-transValue[0], -transValue[1], -transValue[2]);
    
    glGetFloatv(GL_MODELVIEW_MATRIX, currentTransformMat);
    
    for (i = 0; i < numCurrentObjs; i++) {

      glLoadMatrixf(currentTransformMat);
      glMultMatrixf(currentObjList[i]->objMat);
      glGetFloatv(GL_MODELVIEW_MATRIX, currentObjList[i]->objMat);

      applyMtoV(currentObjList[i]->center, currentObjList[i]->center, currentTransformMat); 


    }
    glPopMatrix();

    LdownXold = Xt;
    LdownYold = Yt;

#if DRAWJOINTS
    if (displayPoints == 1) {
      glPushMatrix();
      glLoadMatrixf(currentTransformMat);
      glMultMatrixf(jointModelMat);
      glGetFloatv(GL_MODELVIEW_MATRIX, jointModelMat);
      glPopMatrix();
    }
#endif

    if (numCurrentObjs > 0)
      display();
  }
}

  
void mouseLeftUp(AUX_EVENTREC *mouseEvent) {

  LdownXold = -1.0; 
  LdownYold = -1.0; 

}


void mouseMiddleDown(AUX_EVENTREC *mouseEvent) {

  OBJ* myobject;

  GLfloat Xval, Yval, Xt, Yt;
  int i;

  Xt = (float)(mouseEvent->data[AUX_MOUSEX]);
  Yt = (float)(mouseEvent->data[AUX_MOUSEY]);

  if ((MdownXold == -1.0) && (MdownYold == -1.0)) {
    MdownXold = Xt;
    MdownYold = Yt;
    if (numCurrentObjs != 2) 
      displayPoints = 0;
  }
  else {
    
    Xval = ((Xt - MdownXold)/500.0)*SIZEX;
    Yval = ((MdownYold - Yt)/500.0)*SIZEY;

    glPushMatrix();
    
    glLoadIdentity();
    glTranslatef(Xval, Yval, 0.0);
    glGetFloatv(GL_MODELVIEW_MATRIX, currentTransformMat);
    
    for (i = 0; i < numCurrentObjs; i++) {
      glLoadMatrixf(currentTransformMat);
      glMultMatrixf(currentObjList[i]->objMat);
      glGetFloatv(GL_MODELVIEW_MATRIX, currentObjList[i]->objMat);


      currentObjList[i]->center[0] += Xval;
      currentObjList[i]->center[1] += Yval;

    }    
    glPopMatrix();

    MdownXold = Xt;
    MdownYold = Yt;

#if DRAWJOINTS
    if (displayPoints == 1) {
      glPushMatrix();
      glLoadMatrixf(currentTransformMat);
      glMultMatrixf(jointModelMat);
      glGetFloatv(GL_MODELVIEW_MATRIX, jointModelMat);
      glPopMatrix();
    }
#endif
    if (numCurrentObjs > 0)
      display();
  }
}
    
  
void mouseMiddleUp(AUX_EVENTREC *mouseEvent) {
  
  MdownXold = -1.0; 
  MdownYold = -1.0; 

}    
  
void mouseRightDown(AUX_EVENTREC *mouseEvent) {

  GLfloat Xval, Yval, Xt, Yt;
  int i;

  Xt = (float)(mouseEvent->data[AUX_MOUSEX]);
  Yt = (float)(mouseEvent->data[AUX_MOUSEY]);

  if ((RdownXold == -1.0) && (RdownYold == -1.0)) {    
    RdownXold = Xt;
    RdownYold = Yt;
    if (numCurrentObjs != 2) 
      displayPoints = 0;
  }
  else {

    Xval = ((Xt - RdownXold)/500.0)*SIZEX;
    Yval = ((RdownYold - Yt)/500.0)*SIZEY;

    glPushMatrix();
    
    glLoadIdentity();
    glTranslatef(0.0, 0.0, Yval);
    glGetFloatv(GL_MODELVIEW_MATRIX, currentTransformMat);
    
    for (i = 0; i < numCurrentObjs; i++) {
      glLoadMatrixf(currentTransformMat);
      glMultMatrixf(currentObjList[i]->objMat);
      glGetFloatv(GL_MODELVIEW_MATRIX, currentObjList[i]->objMat);

      currentObjList[i]->center[2] += Yval;

    }    

    glPopMatrix();

    RdownXold = Xt;
    RdownYold = Yt;


#if DRAWJOINTS
    if (displayPoints == 1) {
      glPushMatrix();
      glLoadMatrixf(currentTransformMat);
      glMultMatrixf(jointModelMat);
      glGetFloatv(GL_MODELVIEW_MATRIX, jointModelMat);
      glPopMatrix();
    }
#endif
    if (numCurrentObjs > 0)
      display();
  }
}
    
  
void mouseRightUp(AUX_EVENTREC *mouseEvent) {
  
  RdownXold = -1.0; 
  RdownYold = -1.0; 

}    

void keyUP() {
  
  distThreshold = distThreshold + 0.5;

  printf("Distance Threshold = %f\n", distThreshold);

}

void keyDOWN() {
  
  if (distThreshold >= 1.5) {
    distThreshold = distThreshold - 0.5;
  }

  printf("Distance Threshold = %f\n", distThreshold);

}


#endif /* DISPLAYON */

/****************************************************************************************
                                    I/O
****************************************************************************************/

void freeObj(OBJ* t) {
  free(t->ssArray);
  free(t->sstype);
  free(t->objSeq);
}

int inputObject(int num) {

  char tempname[300];
  int i, imax, count;

  if ((num == 0) && (globalgotone == 1)) {
    freeObj(ObjList[0]);
    free(ObjList[0]);
    globalgotone = 0;
  }

  if ((num == 1) && (globalgottwo == 1)) {
    freeObj(ObjList[1]);
    free(ObjList[1]);
    globalgottwo = 0;
  }

  ObjList[num] = readFile(num);

  if (ObjList[num] != NULL) {
    
    if (newfile[num] == 1) {
      
      if (num == 0)
	printf("\n** Query = %s ", username[0]);
      else 
	printf("\n** Target = %s ", username[1]);
      
      if (chain[num] != ' ') {
	printf("  Chain = %c **\n", chain[num]);
      }
      else {
	printf("**\n");
      }
    }
    else {
      if (num == 0)
	printf("\n** Query = %s **\n", pdbname[0]);
      else 
	printf("\n** Target = %s **\n", pdbname[1]);
    }

    printf("%s\n", ObjList[num]->sstype);
    
    if (num == 0) {
      globalgotone = 1;
      numCurrentObjs = 1;
    }
    else {
      globalgottwo = 1;
      numCurrentObjs = 2;
    }

    currentObjList[num] = ObjList[num];
    displayPoints = 0;

#ifdef DISPLAYON
    if (globalStartDisplay) {
      if (newfile[num] == 1) {
	imax = strlen(username[num]);
	count = 0;
	
	for(i = 0; i < imax; i++) {
	  if (username[num][i] != '/') {	
	    tempname[count] = username[num][i];
	    count++;
	  }
	  else {
	    count = 0;
	  }
	}
	tempname[count] = '\0';	
	tellTcl(num, tempname, 1);    
      }
      else {
	tellTcl(num, pdbname[num], 1);
      }
      display();
    }
#endif    
    return 1;
  }
  else {
    return 0;
  }
}

void setupOutputFile() {

  char outFileStr[300];
  char temp1[300];
  char syscommand[50];

  strcpy(outFileStr, pdbname[0]);
  strcat(outFileStr, "_");
  strcat(outFileStr, pdbname[1]);
  strcat(outFileStr, ".out");

  if (outfile != NULL) {
    fclose(outfile);
  }

  if (strcmp(outputDir, "NONE") == 0) {
    outfile = fopen(outFileStr, "w");
  }

  else {
    strcpy(temp1, outputDir);
    strcat(temp1, "/");
    strcat(temp1, outFileStr);    
    outfile = fopen(temp1, "w");

    if (outfile == NULL) {
      printf("Trying to create directory %s\n", outputDir);
      sprintf(syscommand, "mkdir %s", outputDir);
      system(syscommand);
      outfile = fopen(temp1, "w");
    }
  }

  if (outfile == NULL) {
    printf("Sorry.  Could not open output file %s", outFileStr);
  }      

}

void printOutput() {
  
  setupOutputFile();
  printStructAlignment();
  printAlignmentMatrices();
  fclose(outfile);
  outfile = NULL;	    
}
  

void printStructAlignment() {

  int i,j, tpoint, mycount=0, outcount;
  OBJ *t0, *t1;
  char qlist[MAXLENGTH];
  char alist[MAXLENGTH];
  char tlist[MAXLENGTH];
  int ilen, tlen, previ, prevt;

  if (outfile != NULL) {

    t0 = ObjList[0];
    t1 = ObjList[1];

    for (i = 0; i < t0->numPts; i++) {
      if (t0->pairNums[i] != -1) {
	mycount++;
      }
    }

    if (newfile[0] == 1) {
      fprintf(outfile, "Query: %s  ", username[0]);
      if (chain[0] != ' ') {
	fprintf(outfile, "Chain: %c\n", chain[0]);
      }
      else {
	fprintf(outfile, "\n");
      }
    }
    else {
      fprintf(outfile, "Query: %s\n", pdbname[0]);
    }
    
    if (newfile[1] == 1) {
      fprintf(outfile, "Target: %s  ", username[1]);
      if (chain[1] != ' ') {
	fprintf(outfile, "Chain: %c\n", chain[1]);
      }
      else {
	fprintf(outfile, "\n");
      }
    }
    else {
      fprintf(outfile, "Target: %s\n", pdbname[1]);
    }
    
	
    fprintf(outfile, "RMSD = %f  Num Pre-Core Aligned = %d \n", globalRMSD, globalNumAligned);
    if ((globalComputeCore) || (globalLastAlignment == 1))
      fprintf(outfile, "RMSD core = %f  Num Core Aligned = %d\n", globalCoreRMSD, globalCoreNumAligned);
    if ((t0->ssArrayLength == 0) || (t1->ssArrayLength == 0)) {
      fprintf(outfile, "NOSS\n");
    }
    else {
      fprintf(outfile, "SSOK\n");
    }      
	

    if (globalLastAlignment == 1) {

      outcount = 0;
      previ = -1;
      prevt = -1;

      fprintf(outfile, "\n**********************\n\n");

      for (i = 0; i < t0->numPts; i++) { 
	if (t0->pairNums[i] != -1) {
	  ilen = i - previ - 1;
	  tlen = t0->pairNums[i] - prevt - 1;
	  if (ilen > tlen) {	    
	    for (j = 0; j < ilen; j++) {
	      qlist[outcount] = t0->objSeq[previ + j + 1];
	      alist[outcount] = ' ';
	      
	      if (j < tlen/2) {
		tlist[outcount] = t1->objSeq[prevt + j + 1];
	      }
	      else if ((j >= tlen/2) && (j < tlen/2 + (ilen - tlen))) {
		tlist[outcount] = '-';
	      }
	      else {
		tlist[outcount] = t1->objSeq[prevt + j - (ilen - tlen) + 1];
	      }
	      outcount++;
	    }
	  }
	  else if (ilen < tlen) {
	    for (j = 0; j < tlen; j++) {
	      tlist[outcount] = t1->objSeq[prevt + j + 1];
	      alist[outcount] = ' ';
	      
	      if (j < ilen/2) {
		qlist[outcount] = t0->objSeq[previ + j + 1];
	      }
	      else if ((j >= ilen/2) && (j < ilen/2 + (tlen - ilen))) {
		qlist[outcount] = '-';
	      }
	      else {
		qlist[outcount] = t0->objSeq[previ + j - (tlen - ilen) + 1];
	      }
	      outcount++;
	    }
	  }
	  else {
	    for (j = 0; j < ilen; j++) {
	      qlist[outcount] = t0->objSeq[previ + j + 1];
	      alist[outcount] = ' ';
	      tlist[outcount] = t1->objSeq[prevt + j + 1];
	      outcount++;
	    }
	  }

	  qlist[outcount] = t0->objSeq[i];
	  alist[outcount] = '|';
	  tlist[outcount] = t1->objSeq[t0->pairNums[i]];
	  outcount++;

	  previ = i;
	  prevt = t0->pairNums[i];

	}
      }

      ilen = t0->numPts - previ - 1;
      tlen = t1->numPts - prevt - 1;

      j = (ilen > tlen)?ilen:tlen;

      for (i = 1; i <= j; i++) {
	alist[outcount] = ' ';
	if (previ + i < t0->numPts) {
	  qlist[outcount] = t0->objSeq[previ+i];
	}
	else {
	  qlist[outcount] = '-';
	}
	if (prevt + i < t1->numPts) {
	  tlist[outcount] = t1->objSeq[prevt+i];
	}
	else {
	  tlist[outcount] = '-';
	}
	outcount++;
      }

      j = 0;

      while (j < outcount) {
	fprintf(outfile, "Query \t");
	for (i = j; ((i < j + 50) && (i < outcount)); i++) {
	  if (i%10 == 0) 
	    fprintf(outfile, " ");
	  fprintf(outfile, "%c", qlist[i]);
	}
	fprintf(outfile, "\n");
	
	fprintf(outfile, "      \t");
	for (i = j; ((i < j + 50) && (i < outcount)); i++) {
	  if (i%10 == 0) 
	    fprintf(outfile, " ");
	  fprintf(outfile, "%c", alist[i]);
	}
	fprintf(outfile, "\n");
	
	fprintf(outfile, "Target\t");
	for (i = j; ((i < j + 50) && (i < outcount)); i++) {
	  if (i%10 == 0) 
	    fprintf(outfile, " ");
	  fprintf(outfile, "%c", tlist[i]);
	}
	fprintf(outfile, "\n");
	j = j + 50;
	fprintf(outfile, "\n");
      }

      fprintf(outfile, "\nNote that only residues matched with '|' correspond to structurally\naligned residue pairs (see following list).\n");

    }  
	

    fprintf(outfile, "\n**********************\n");
    if (globalLastAlignment == 0) {
      fprintf(outfile, "\nNum Pre-Core Aligned = %d\n", mycount);
    }
    else if (globalLastAlignment == 1) {
      fprintf(outfile, "\nNum Core Aligned = %d\n", mycount);
    }
    for (i = 0; i < t0->numPts; i++) {
      if (t0->pairNums[i] != -1) {
	tpoint = t0->pairNums[i];
	fprintf(outfile,"%d %c\t", t0->aaNums[i], t0->objSeq[i]);
	fprintf(outfile,"%d %c\t%f\n", t1->aaNums[tpoint], t1->objSeq[tpoint], t0->pairDists[i]);

      }
    }
    
    fprintf(outfile, "\n\n**********************\n\n");
    
    printf("Alignment saved\n");    		    
  }

}


void printAlignmentMatrices() {
  
  char *str0, *str1;

  if (outfile != NULL) {

    if (newfile[0] == 1) {
      str0 = username[0];
    }
    else {
      str0 = pdbname[0];
    }
    
    if (newfile[1] == 1) {
      str1 = username[1];
    }
    else {
      str1 = pdbname[1];
    }
    

    fprintf(outfile, "\nTransformation Matrix for query %s:", str0);
    printMat(ObjList[0]->objMat);

    fprintf(outfile, "\nCenter of mass for %s: \n", str0);
    fprintf(outfile, "%f  %f  %f\n", ObjList[0]->center[0], ObjList[0]->center[1], ObjList[0]->center[2]);

    fprintf(outfile, "\nTransformation Matrix for target %s:", str1);
    printMat(ObjList[1]->objMat);

    fprintf(outfile, "\nCenter of mass for %s: \n", str1);
    fprintf(outfile, "%f  %f  %f\n", ObjList[1]->center[0], ObjList[1]->center[1], ObjList[1]->center[2]);

    printf("Matrices saved\n\n");
    
  }


}


void loadAlignmentMatrices() {
  
  FILE *matrixFile;
  char matrixFileName[50];
  GLfloat temp;
  char tempstr[10];
  int i;
  int tempint;

  printf("Enter Matrix File Name > ");

  scanf("%s", matrixFileName);

  if ((matrixFile = fopen(matrixFileName, "r")) == NULL) {
    printf("Cannot open matrix file %s\n", matrixFileName);
  }

  else {  

    fscanf(matrixFile, "%s", tempstr);
    fscanf(matrixFile, "%s", tempstr);
    fscanf(matrixFile, "%f %d", &temp, &tempint);

    fscanf(matrixFile, "%s", tempstr);
    fscanf(matrixFile, "%s", tempstr);
    fscanf(matrixFile, "%s", tempstr);
    fscanf(matrixFile, "%f %d", &temp, &tempint);
    fscanf(matrixFile, "%s", tempstr);
    
    for (i = 0; i < 16; i++) {
      fscanf(matrixFile, "%f", &(ObjList[0]->objMat[i]));
    }
    
    fscanf(matrixFile, "%f %f %f", &(ObjList[0]->center[0]), &(ObjList[0]->center[1]), &(ObjList[0]->center[2]));
    
    for (i = 0; i < 16; i++) {
      fscanf(matrixFile, "%f", &(ObjList[1]->objMat[i]));
    }

    fscanf(matrixFile, "%f %f %f", &(ObjList[1]->center[0]), &(ObjList[1]->center[1]), &(ObjList[1]->center[2]));
        
    displayPoints = 0;
#ifdef DISPLAYON
    if (globalStartDisplay) {
      display();
    }
#endif
    fclose(matrixFile);
    
  }

}


/****************************************************************************************
                                    Alignment Functions
				    Align using vectors (step 1)
****************************************************************************************/

float alignUsingVectors(OBJ* query, OBJ* target, int iterations, float epsilon, int full) {

  int i,j, converged, iter, tnum, count = 0;
  Vector translation;
  Matrix mat;
  GLfloat m[16];
  OBJ tempObj1, tempObj0;
  GLfloat tdist, rmsd, diffrmsd, oldrmsd = 10000.0;
  GLfloat tempVec[3];

  for (i = 0; i < target->numPts; i++) {
    target->pairNums[i] = -1;
  }
  for (i = 0; i < query->numPts; i++) {
    query->pairNums[i] = -1;
  }

  tempObj0.objSeq = query->objSeq;
  tempObj0.sstype = query->sstype;
  tempObj0.numPts = query->numPts;
  tempObj0.ssArrayLength = query->ssArrayLength;
  tempObj0.ssArray = query->ssArray;
  applyMat(&tempObj0, query, query->objMat);
  
  tempObj1.objSeq = target->objSeq;
  tempObj1.sstype = target->sstype;
  tempObj1.numPts = target->numPts;
  tempObj1.ssArrayLength = target->ssArrayLength;
  tempObj1.ssArray = target->ssArray;
  applyMat(&tempObj1, target, target->objMat);
    
  converged = 0;

  for (iter = 0; iter < iterations; iter++) {

    init_pairs();
    
    initsequences_ss(&tempObj0, &tempObj1);
    align_ss(full);
    report_ss(full);

    if (globalCalledAddPair >= 4) {

      match_pairs (mat, translation);
      
      for (i = 0; i < 4; i++) {
	for (j = 0; j < 4; j++) {
	  m[i*4 + j] = mat[i][j];
	}
      }

      myMatMult(target->objMat, m);
      applyMtoV(target->center, target->center, m);

      applyMat(&tempObj1, target, target->objMat);

      count = 0;
      rmsd = 0.0;

      for (i = 0; i < tempObj1.numPts; i++) {
	tdist = findNearest(&tempObj0, tempObj1.points[i], tempObj1.sstype[i], tempVec, &tnum);
	
	if ((tdist <= distThreshold) || (thresholdOFF)) {
	  rmsd += tdist*tdist;
	  count++;
	}
	  
      }
      
      rmsd = sqrt((rmsd)/((GLfloat)(count)));
      globalNumAligned = count;

      diffrmsd = oldrmsd - rmsd;

      if (diffrmsd < 0) 
	diffrmsd *= -1.0;
      
      if ((diffrmsd < epsilon) && (iter < iterations - 1))  {
	iter = iterations - 1;
	converged = 1;
      }
      
      oldrmsd = rmsd;

      
    }
    else {
      iter = iterations;
      rmsd = 0.0;
      globalNumAligned = 0;
    }  
  }

  
  if (globalNumAligned != 0) {
    
    count = 0;
    rmsd = 0.0;

    for (i = 0; i < tempObj1.numPts; i++) {
      tdist = findNearest(&tempObj0, tempObj1.points[i], tempObj1.sstype[i], tempVec, &tnum);
      
      if ((tdist <= distThreshold) || (thresholdOFF)) {
	target->pairNums[i] = tnum;
	target->pairDists[i] = tdist;


	if (query->pairNums[tnum] != -1) {
	  if (tdist < query->pairDists[tnum]) {	    
	    rmsd -= (query->pairDists[tnum])*(query->pairDists[tnum]);
	    query->pairNums[tnum] = i;	    
	    query->pairDists[tnum] = tdist;
	    rmsd += tdist*tdist;
	  }
	}
	else {
	  query->pairNums[tnum] = i;	    
	  query->pairDists[tnum] = tdist;
	  rmsd += tdist*tdist;
	  count++;
	}	  
      }
    }

    rmsd = sqrt((rmsd)/((GLfloat)(count)));
    globalNumAligned = count;

  }

  
  if ((converged == 0) && (iterations == MAXITERATIONS)) {
    printf("Did not converge\n");
  }

  return rmsd;

}




void alignUsingVectorsSingle() {

  OBJ *t0, *t1;
  float rmsd;
  int i;

  if ((ObjList[0]->ssArrayLength > 0) && (ObjList[1]->ssArrayLength > 0)) {

    t0 = ObjList[0];
    t1 = ObjList[1];
    
    for (i = 0; i < 5; i++) {
      rmsd = alignUsingVectors(t0, t1, 1, 0.0, 1);
      
      globalRMSD = rmsd;
      
      printf("RMSD = %f %d\n", rmsd, globalNumAligned);

    }
#ifdef DISPLAYON
    if (globalStartDisplay) {
#if DRAWJOINTS
      createJointList();
#endif
      display();
    }
#endif
  }

}


void alignUsingVectorsFull() {

  OBJ *t0, *t1;
  float rmsd;

  if ((ObjList[0]->ssArrayLength > 0) && (ObjList[1]->ssArrayLength > 0)) {

    t0 = ObjList[0];
    t1 = ObjList[1];

    rmsd = alignUsingVectors(t0, t1, MAXITERATIONS, MAXDIFFERENCE, 1);

    globalRMSD = rmsd;
    
    printf("RMSD = %f %d\n", rmsd, globalNumAligned);

#ifdef DISPLAYON
    if (globalStartDisplay) {

#if DRAWJOINTS
      createJointList();
#endif
      display();
    }
#endif
  }


}


void initAlignUsingVectorsSingle() {

  OBJ *t0, *t1;
  float rmsd;

  if ((ObjList[0]->ssArrayLength > 0) && (ObjList[1]->ssArrayLength > 0)) {

    t0 = ObjList[0];
    t1 = ObjList[1];

    rmsd = alignUsingVectors(t0, t1, 1, 0.0, 0);
    
    globalRMSD = rmsd;
    
    printf("RMSD = %f %d\n", rmsd, globalNumAligned);
   
#ifdef DISPLAYON
    if (globalStartDisplay) {   
#if DRAWJOINTS
      createJointList();
#endif
      display();
    }
#endif
  }
}


void initAlignUsingVectorsFull() {

  OBJ *t0, *t1;
  float rmsd;

  if ((ObjList[0]->ssArrayLength > 0) && (ObjList[1]->ssArrayLength > 0)) {

    t0 = ObjList[0];
    t1 = ObjList[1];
    
    rmsd = alignUsingVectors(t0, t1, MAXITERATIONS, MAXDIFFERENCE, 0);

    globalRMSD = rmsd;
    
    printf("RMSD = %f %d\n", rmsd, globalNumAligned);
    
#ifdef DISPLAYON
    if (globalStartDisplay) {
#if DRAWJOINTS
      createJointList();
#endif
      display();
    }
#endif
  }

}


/****************************************************************************************
                                    Alignment Functions
				    Minimize RMSD (step 2)
****************************************************************************************/


float structAlign(OBJ* query, OBJ* target, int iterations, float epsilon) {

  int i, j, iter, converged, count, tnum;
  Vector translation;
  Matrix mat;
  GLfloat m[16];
  GLfloat tempVec[3];
  GLfloat tdist;
  GLfloat rmsd = 0.0, oldrmsd = 1000.0, diffrmsd;
  OBJ tempObj1, tempObj0;


  tempObj0.numPts = query->numPts;
  tempObj0.sstype = query->sstype;
  applyMat(&tempObj0, query, query->objMat);

  tempObj1.numPts = target->numPts;
  tempObj1.sstype = target->sstype;

  converged = 0;

  for (i = 0; i < target->numPts; i++) {
    target->pairNums[i] = -1;
  }
  for (i = 0; i < query->numPts; i++) {
    query->pairNums[i] = -1;
  }

  rmsd = 0.0;
  
  applyMat(&tempObj1, target, target->objMat);
 
  init_pairs();
  
  count = 0;
  
  for (i = 0; i < tempObj1.numPts; i++) {
    tdist = findNearest(&tempObj0, tempObj1.points[i], tempObj1.sstype[i], tempVec, &tnum);
    
    if ((tdist <= distThreshold) || (thresholdOFF)) {
      add_pair(tempVec, tempObj1.points[i], 1);
      rmsd += tdist*tdist;
      count++;
    }
  }


  for (iter = 0; iter < iterations; iter++) {
    
    if (count >= 4) { 

      match_pairs (mat, translation);
      
      for (i = 0; i < 4; i++) {
	for (j = 0; j < 4; j++) {
	  m[i*4 + j] = mat[i][j];
	}
      }
      
      myMatMult(target->objMat, m);
      applyMtoV(target->center, target->center, m);
           
      rmsd = 0.0;
      
      applyMat(&tempObj1, target, target->objMat);
            
      init_pairs();
    
      count = 0;
      
      for (i = 0; i < tempObj1.numPts; i++) {
	tdist = findNearest(&tempObj0, tempObj1.points[i], tempObj1.sstype[i], tempVec, &tnum);
	
	if ((tdist <= distThreshold) || (thresholdOFF)) {
	  add_pair(tempVec, tempObj1.points[i], 1);
	  rmsd += tdist*tdist;
	  count++;
	}
      }
      
      rmsd = sqrt((rmsd)/((GLfloat)(count)));
      globalNumAligned = count;
      
      diffrmsd = oldrmsd - rmsd;
      
      if (diffrmsd < 0) 
	diffrmsd *= -1.0;
      
      if ((diffrmsd < epsilon) && (iter < iterations - 1))  {
	iter = iterations - 1;
	converged = 1;
      }
    
      oldrmsd = rmsd;
    
    }
    else {

      printf("There are less than four atoms correctly aligned\n");      
      globalNumAligned = 0;
      rmsd = 0.0;
      iter = iterations - 1;

    }
  }


  if (globalNumAligned != 0) {
    
    count = 0;
    rmsd = 0.0;

    for (i = 0; i < tempObj1.numPts; i++) {
      tdist = findNearest(&tempObj0, tempObj1.points[i], tempObj1.sstype[i], tempVec, &tnum);
      
      if ((tdist <= distThreshold) || (thresholdOFF)) {
	target->pairNums[i] = tnum;
	target->pairDists[i] = tdist;


	if (query->pairNums[tnum] != -1) {
	  if (tdist < query->pairDists[tnum]) {	    
	    rmsd -= (query->pairDists[tnum])*(query->pairDists[tnum]);
	    query->pairNums[tnum] = i;	    
	    query->pairDists[tnum] = tdist;
	    rmsd += tdist*tdist;
	  }
	}
	else {
	  query->pairNums[tnum] = i;	    
	  query->pairDists[tnum] = tdist;
	  rmsd += tdist*tdist;
	  count++;
	}	  
      }
    }

    rmsd = sqrt((rmsd)/((GLfloat)(count)));
    globalNumAligned = count;

  }


  if ((converged == 0) && (iterations == MAXITERATIONS)) {
    printf("Did not converge\n");
  }



  globalLastAlignment = 0;

  return rmsd;
  
}


void structAlignSingle() {

  OBJ *t0, *t1;
  float rmsd;
  int i;

  t0 = ObjList[0];
  t1 = ObjList[1];
  
  for (i = 0; i < 10; i++) {
    rmsd = structAlign(t0, t1, 1, 0.0);
    
    globalRMSD = rmsd;
    
    printf("RMSD = %f %d\n", rmsd, globalNumAligned);
    
#ifdef DISPLAYON
    if (globalStartDisplay) {
      display();
    }
#endif
  }

#ifdef DISPLAYON
  if (globalStartDisplay) {
#if DRAWJOINTS
    createJointList();
#endif
    display();
  }
#endif

}


void structAlignFull() {

  OBJ *t0, *t1;
  float rmsd;

  t0 = ObjList[0];
  t1 = ObjList[1];
  
  rmsd = structAlign(t0, t1, MAXITERATIONS, MAXDIFFERENCE);
  
  globalRMSD = rmsd;
  
  printf("RMSD = %f %d\n", rmsd, globalNumAligned);
  
#ifdef DISPLAYON
  if (globalStartDisplay) { 
#if DRAWJOINTS
    createJointList();
#endif
    display();
  }
#endif
}

/****************************************************************************************
                                    Alignment Functions
				    Core Align (step 3)
****************************************************************************************/


void findBestSequence(int *a, float *dist, int start, int end) {

  int i, j;
  int myscore, mypointer, maxscore, maxindex;
  int scorearray[MAXLENGTH];
  int pointarray[MAXLENGTH];

  maxscore = -100;
  maxindex = -1;

  for (i = 0; i < end; i++) {
    scorearray[i] = -1;
    pointarray[i] = -1;
    
    if (a[i] != -1) {
      myscore = 1;
      mypointer = -1;
      for (j = i-1; j >=0; j--) {  
	if ((a[j] != -1) && (a[j] < a[i])) {
	  if (scorearray[j] + 1 > myscore) {	
	    myscore = scorearray[j] + 1;
	    mypointer = j;
	  }
	  else if (scorearray[j] + 1 == myscore) {
	    if (dist[j] < dist[mypointer]) {
	      mypointer = j;
	    }
	  }
	}
      }  
      scorearray[i] = myscore;
      pointarray[i] = mypointer;
      if (myscore > maxscore) {
	maxscore = myscore;
	maxindex = i;
      }
    }

  }

  while (maxindex != -1) {
    scorearray[maxindex] = 0;
    maxindex = pointarray[maxindex];
  }

  for (i = 0; i < end; i++) {
    if (scorearray[i] != 0) {
      a[i] = -1;
    }
  }
}


float refineAlign(OBJ* query, OBJ* target, int iterations, float epsilon, int doss) {

  OBJ tempObj1, tempObj0;
  GLfloat tempVec[3];
  GLfloat tdist;
  int i, j, tnum;
  float rmsd = 0.0, oldrmsd = 1000.0, diffrmsd;
  int iter, converged = 0;

  Vector translation;
  Matrix mat;
  GLfloat m[16];
  
  float distArray1[MAXLENGTH];
  int correctArray1[MAXLENGTH];
  int len1;

  int posArray0[MAXLENGTH];
  float distArray0[MAXLENGTH];
  int numcorrect = 0;
  int len0;

  int totalCount = 0;
  float myrmsd = 0.0;

  for (i = 0; i < target->numPts; i++) {
    target->pairNums[i] = -1;
  }  

  for (i = 0; i < query->numPts; i++) {
    query->pairNums[i] = -1;
  }  
  
  tempObj1.numPts = target->numPts;
  tempObj0.numPts = query->numPts;

  tempObj1.sstype = target->sstype;
  tempObj0.sstype = query->sstype;
  
  applyMat(&tempObj0, query, query->objMat);
  
  
  len1 = target->numPts;
  len0 = query->numPts;

  init_pairs();
  
  numcorrect = 0;
  
  totalCount = 0;
  myrmsd = 0.0;
  
  applyMat(&tempObj1, target, target->objMat);    
  
  for (i = 0; i < len1; i++) {    
    distArray1[i] = 1000.0;;
    correctArray1[i] = -1;
  }    
    
  for (i = 0; i < len0; i++) {    
    tdist = findNearest(&tempObj1, tempObj0.points[i], query->sstype[i], tempVec, &tnum);
    posArray0[i] = tnum;    
    distArray0[i] = tdist;
  }
              

  for (i = 0; i < len0; i++) {
    if ((distArray0[i] < distArray1[posArray0[i]]) && ((distArray0[i] < distThreshold) || (thresholdOFF))) {
      correctArray1[posArray0[i]] = i;
      distArray1[posArray0[i]] = distArray0[i];
    }
  }


  findBestSequence(correctArray1, distArray1, 0, len1);


  for (i = 0; i < len1; i++) {
    if (correctArray1[i] != -1) {
      add_pair(tempObj0.points[correctArray1[i]], tempObj1.points[i], 1);
      totalCount++;
      myrmsd += (distArray1[i])*(distArray1[i]);
    }
  }

  
  for (iter = 0; iter < iterations; iter++) {

    if (totalCount >= 4) {
      
      match_pairs (mat, translation);
    
      for (i = 0; i < 4; i++) {
	for (j = 0; j < 4; j++) {
	  m[i*4 + j] = mat[i][j];
	}
      }
      
      
      myMatMult(target->objMat, m);
      applyMtoV(target->center, target->center, m);      


      init_pairs();
      
      numcorrect = 0;
      
      totalCount = 0;
      myrmsd = 0.0;
      
      applyMat(&tempObj1, target, target->objMat);    
      
      for (i = 0; i < len1; i++) {    
	distArray1[i] = 1000.0;
	correctArray1[i] = -1;
      }    
      
      for (i = 0; i < query->numPts; i++) {    
	tdist = findNearest(&tempObj1, tempObj0.points[i], query->sstype[i], tempVec, &tnum);
	posArray0[i] = tnum;    
	distArray0[i] = tdist;
      }
            
      
      for (i = 0; i < query->numPts; i++) {
	if ((distArray0[i] < distArray1[posArray0[i]]) && ((distArray0[i] < distThreshold) || (thresholdOFF))) {
	  correctArray1[posArray0[i]] = i;
	  distArray1[posArray0[i]] = distArray0[i];
	}
      }
      
      findBestSequence(correctArray1, distArray1, 0, len1);

      for (i = 0; i < len1; i++) {
	if (correctArray1[i] != -1) {
	  add_pair(tempObj0.points[correctArray1[i]], tempObj1.points[i], 1);
	  totalCount++;
	  myrmsd += (distArray1[i])*(distArray1[i]);
	}
      }

      rmsd = sqrt((myrmsd)/((GLfloat)(totalCount)));
      diffrmsd = oldrmsd - rmsd;
      
      if (diffrmsd < 0) 
	diffrmsd *= -1.0;
      
      if ((diffrmsd < epsilon) && (iter < iterations - 1))  {
	iter = iterations - 1;
	converged = 1;
      }
      
      oldrmsd = rmsd;

    }
    else {

      printf("There are less than four core atoms correctly aligned\n");      
      totalCount = 0;
      rmsd = 0.0;
      iter = iterations - 1;

    }

  }

  if ((converged == 0) && (iterations == MAXITERATIONS)) {
    printf("Refinealign did not converge\n");
  }

  if (totalCount > 0) {
    for (i = 0; i < len1; i++) {
      if (correctArray1[i] != -1) {
	target->pairNums[i] = correctArray1[i];
	target->pairDists[i] = distArray1[i];


	if (query->pairNums[correctArray1[i]] != -1) {
	  if (distArray1[i] < query->pairDists[correctArray1[i]]) {
	    query->pairNums[correctArray1[i]] = i;
	    query->pairDists[correctArray1[i]] = distArray1[i];
	  }
	}
	else {
	  query->pairNums[correctArray1[i]] = i;
	  query->pairDists[correctArray1[i]] = distArray1[i];
	}	  
      }
    }
  }

  globalCoreNumAligned = totalCount;

  globalLastAlignment = 1;

  return rmsd;

}


void coreAlignSingle() {
  OBJ *t0, *t1;
  float rmsd;
  int i;

  t0 = ObjList[0];
  t1 = ObjList[1];
  
  for (i = 0; i < 5; i++) {
    rmsd = refineAlign(t0, t1, 1, 0.0, 1);
    
    globalCoreRMSD = rmsd;
    
    printf("Core RMSD = %f %d\n", globalCoreRMSD, globalCoreNumAligned);
    
#ifdef DISPLAYON
    if (globalStartDisplay) {
      display();
    }
#endif
  }

#ifdef DISPLAYON
  if (globalStartDisplay) {
#if DRAWJOINTS
    createJointList();
#endif
    display();
  }
#endif  
}


void coreAlignFull() {

  OBJ *t0, *t1;
  float rmsd;

  t0 = ObjList[0];
  t1 = ObjList[1];
  
  rmsd = refineAlign(t0, t1, MAXITERATIONS, MAXDIFFERENCE, 1);
  
  globalCoreRMSD = rmsd;
  
  printf("Core RMSD = %f %d\n", globalCoreRMSD, globalCoreNumAligned);

#ifdef DISPLAYON
  if (globalStartDisplay) {    
#if DRAWJOINTS
    createJointList();
#endif
    display();
  }
#endif
}

  
/****************************************************************************************
                                    Alignment Functions
				    Sequence Align (unused)
****************************************************************************************/

void sequenceAlign() {

  int i,j;
  Vector translation;
  Matrix mat;
  GLfloat m[16];
  OBJ tempObj1, tempObj0;
  GLfloat tdist;

  init_pairs();
  tempObj1.objSeq = ObjList[0]->objSeq;
  tempObj1.sstype = ObjList[0]->sstype;
  tempObj1.numPts = ObjList[0]->numPts;
  applyMat(&tempObj1, ObjList[0], ObjList[0]->objMat);
  
  tempObj0.objSeq = ObjList[1]->objSeq;
  tempObj0.sstype = ObjList[1]->sstype;
  tempObj0.numPts = ObjList[1]->numPts;
  applyMat(&tempObj0, ObjList[1], ObjList[1]->objMat);
  
  
  initsequences(&tempObj0, &tempObj1);
  align();
  report();
  
  match_pairs (mat, translation);
  
  for (i = 0; i < 4; i++) {
    for (j = 0; j < 4; j++) {
      m[i*4 + j] = mat[i][j];
    }
  }
    
  myMatMult(currentObjList[0]->objMat, m);
  applyMtoV(currentObjList[0]->center, currentObjList[0]->center, m);    
  applyMat(&tempObj1, ObjList[0], ObjList[0]->objMat);
  
  tdist = computeRMSD(&tempObj0, &tempObj1);
  
  globalRMSD = tdist;
  
  printf("RMSD = %f %d\n", tdist, globalNumAligned);
  
  displayPoints = 0;
#ifdef DISPLAYON
  if (globalStartDisplay) {
    display();
  }
#endif
}


void quitFunction() {

  exit(0);

}

void doAllAlignments() {

  int oldt;
  FILE *searchfile;
  char myprot[50];
  float tempcentervector[3];
  float tempobjectmatrix[16];
  int i, done;
  int wordlen;
  char line[200];
  char tempstr[300];

  if (strcmp(outputDir, "NONE") == 0) {
      searchfile = fopen("search.out", "w");
  }
  else {
    strcpy(tempstr, outputDir);
    strcat(tempstr, "/search.out");
    
    searchfile = fopen(tempstr, "w");
    if (searchfile == NULL) {
      printf("Trying to create directory %s\n", outputDir);
      sprintf(line, "mkdir %s", outputDir);
      system(line);
      searchfile = fopen(tempstr, "w");
    }
  }
  
  if (searchfile == NULL) {
    printf("Sorry.  Could not open output file locksearch.out\n");
  }

  else {

    if (globalSearchDatabase == 0) {
      
      if ((ObjList[0]->ssArrayLength > 0) && (ObjList[1]->ssArrayLength > 0)) {
	initAlignUsingVectorsSingle();
	alignUsingVectorsFull();
      }
      
      if (globalNumAligned == 0) {
	oldt = thresholdOFF;
	thresholdOFF = 1;
	structAlignSingle();
	thresholdOFF = oldt;
      }
      
      structAlignFull();
      if (globalComputeCore) {
	coreAlignFull();
	structAlignFull();
	coreAlignFull();
      }
      
      
      printOutput();

      if (globalComputeCore) {
	fprintf(searchfile, "%s\t%d\t%5.3f\t%s\n", pdbname[1], globalCoreNumAligned, globalCoreRMSD, ObjList[1]->header);
      }
      else {
	fprintf(searchfile, "%s\t%d\t%5.3f\t%s\n", pdbname[1], globalNumAligned, globalRMSD, ObjList[1]->header);	  
      }
    }
    
    else if ((globalSearchDatabase == 1) && (globalDatabaseFile != NULL)) {
      
      rewind(globalDatabaseFile);
      
      for (i = 0; i < 3; i++) {
	tempcentervector[i] = ObjList[0]->center[i];
      }
      
      for (i = 0; i < 16; i++) {
	tempobjectmatrix[i] = ObjList[0]->objMat[i];
      }
   
      done = 0;

      while ((done == 0) && (fgets(line, 200, globalDatabaseFile))) {
	sscanf(line, "%s", myprot);
	
	if (strcmp(myprot, "LOCK_LIST") == 0) {	
	  done = 1;
	}
      }
   
      while ((fscanf(globalDatabaseFile, "%s", myprot)) != EOF) {
	
	for (i = 0; i < 3; i++) {
	  ObjList[0]->center[i] = tempcentervector[i];
	}
	
	for (i = 0; i < 16; i++) {
	  ObjList[0]->objMat[i] = tempobjectmatrix[i];
	}
	
	globalCoreNumAligned = 0;
	globalNumAligned = 0;
	globalCoreRMSD = 0.0;
	globalRMSD = 0.0;

	strcpy(username[1], myprot);
	chain[1] = ' ';

	if (inputObject(1)) {
	  if ((ObjList[0]->ssArrayLength > 0) && (ObjList[1]->ssArrayLength > 0)) {	    
	    
	    initAlignUsingVectorsSingle();
	    alignUsingVectorsFull();
	    
	    
	    if (globalNumAligned == 0) {
	      oldt = thresholdOFF;
	      thresholdOFF = 1;
	      structAlignSingle();
	      thresholdOFF = oldt;
	    }
	    
	    structAlignFull();
	    if (globalComputeCore) {
	      coreAlignFull();
	      structAlignFull();
	      coreAlignFull();
	    }

	    printOutput();

	    if (globalComputeCore) {
	      fprintf(searchfile, "%s\t%d\t%5.3f\t%s\n", pdbname[1], globalCoreNumAligned, globalCoreRMSD, ObjList[1]->header);
	    }
	    else {
	      fprintf(searchfile, "%s\t%d\t%5.3f\t%s\n", pdbname[1], globalNumAligned, globalRMSD, ObjList[1]->header);		
	    }
	  }

	}  
	
      }

    }
    fclose(searchfile);
  }
  
}


void main(int argc, char** argv) {
  char str[100];

  freopen("/dev/null", "w", stderr);

  mypid = getpid();

  lockdir = getenv("LOCKDIR");
  if (lockdir == NULL) {
    lockdir = (char*)(malloc((sizeof(char))*200));
    strcpy(lockdir, _LOCKDIR_);
  }

  dsspprogram = getenv("DSSPPROGRAM");
  if (dsspprogram == NULL) {
    dsspprogram = (char*)(malloc((sizeof(char))*200));
    strcpy(dsspprogram, _DSSPPROGRAM_);
  }

  checkopts(argc, argv);

#ifndef DISPLAYON
  globalStartDisplay = 0;
#endif

  if ((strcmp(username[0], "NONE") == 0) || (strcmp(username[1], "NONE") == 0)) {
    printf("\n\nPlease specify a query and target.\n");
    usageError();
  }

#ifdef DISPLAYON
  if (globalStartDisplay) {
    addReshapeFunc(myReshape);
    setBindings();
    
    InitPosition(0, 0, 500, 500);
    My_Tk_Init(1, argv, Tcl_AppInit);        
    setupGLWindow("LOCK");             
    init();

    strcpy(str, lockdir);
    strcat(str, "lock.tcl");
    runTclFile(str);
  }
#endif

  if ((inputObject(0)) && ((inputObject(1)) || (globalSearchDatabase))) {

    if (globalStartDisplay) {      
#ifdef DISPLAYON
      jointModelNumber = glGenLists(1);
      Tk_MainLoop();       
#endif
    }    
    else {
      doAllAlignments();
    }

  }

} 
